/* mpir.h - include file for the MPIR implementation of MPI */

#include <stdio.h>

#ifndef _MPIR_INCLUDE
#define _MPIR_INCLUDE

#include "mpid.h"

/*****************************************************************************
*                           MPIR DATA STRUCTURES                             *
*****************************************************************************/

/* Should switch these to MPIRMALLOC, MPIRFREE, MPIRNEW */
#ifndef MALLOC
#define MALLOC(a)    malloc((unsigned)(a))
#define FREE(a)      free((char *)(a))
#define NEW(a)    (a *)MALLOC(sizeof(a))
#endif

typedef struct _MPIR_DATATYPE {
    MPIR_NODETYPE dte_type;	/* type of datatype element this is */
    MPIR_BOOL committed;	/* whether committed or not */
    int extent;			/* extent of this datatype */
    int use_count;		/* nodes depending on this node */
    struct _MPIR_DATATYPE *old_type; /* type this type is built out of, if 1 */
    int count;			/* replication count */
    int stride;			/* stride, for VECTOR and HVECTOR types */
    int blocklen;		/* blocklen, for VECTOR and HVECTOR types */
    int *indices;		/* array of indices, for (H)INDEXED */
    int *blocklens;		/* array of blocklens for (H)INDEXED */
    struct _MPIR_DATATYPE **old_types;  /* array of types, for STRUCT */
} MPIR_DATATYPE;

typedef enum {
    MPIR_SEND,
    MPIR_RECV
} MPIR_OPTYPE;

typedef struct {
    MPIR_OPTYPE handle_type;	/* send or receive */
    int dest;			/* destination process for message */
    int         tag;		/* tag */
    MPIR_COMMUNICATOR *comm;	/* communicator */
    MPI_MODE mode;		/* mode: STANDARD, READY, SYNC */
    MPIR_BOOL completed;	/* whether operation is completed or not */
    MPI_PERSISTENCE persistence; /* persistence: PERSISTENT or EPHEMERAL */
    void *bufadd;		/* address of buffer */
    MPIR_DATATYPE *datatype;	/* basic or derived datatype */
    
    MPID_SHANDLE dev_shandle;	/* device's version of send handle */
} MPIR_SHANDLE;

typedef struct {
    MPIR_OPTYPE handle_type;	/* send or receive */
    int source;			/* source process message */
    int 	tag;		/* tag */
    MPIR_COMMUNICATOR *comm;	/* communicator */
    MPI_PERSISTENCE persistence; /* persistence: PERSISTENT or EPHEMERAL */
    void *bufadd;		/* address of buffer */
    MPIR_DATATYPE *datatype;	/* basic or derived datatype */
				/* next few fields hold status info */
    void *p;			/* Pointer to unexpected data */
    int  len;			/* Length of this data ? */
    int  count;			/* number of items actually read */
    MPIR_BOOL  completed;	/* Whether this receive has been completed */
    MPID_RHANDLE dev_rhandle;	/* device's version of recv handle */
} MPIR_RHANDLE;

#define MPIR_HANDLES_DEFINED

typedef union {
    MPIR_OPTYPE type;		/* send or receive */
    MPIR_SHANDLE shandle;
    MPIR_RHANDLE rhandle;
} MPIR_HANDLE;

typedef enum {
    MPIR_QSHANDLE,
    MPIR_QRHANDLE
} MPIR_QEL_TYPE;

typedef struct _MPIR_QEL {	/* queue elements */
    void *ptr;			/* queues can contain anything  */
    MPIR_QEL_TYPE qel_type;	/* but we may need to know what */
    struct _MPIR_QEL *prev;	/* previous queue element */
    struct _MPIR_QEL *next;	/* next queue element */
} MPIR_QEL;

typedef struct {		/* header for queues of things like handles */
    MPIR_QEL *first;		/* first queue element */
    MPIR_QEL *last;		/* last queue element */
    int      currlen;		/* current length */
    int      maxlen;		/* maximum since beginning of run */
} MPIR_QHDR;

typedef struct _MPIR_FDTEL {	/* flat datatype elements */
    MPIR_NODETYPE      type;	/* one of the elementary data types */
    int                disp;	/* displacement */
    struct _MPIR_FDTEL *next;	/* pointer to next element */
} MPIR_FDTEL;

#define MPIR_MALLOC(size)            malloc(size)
#define MPIR_dump_datatype(datatype) MPIR_dump_dte(datatype,0)

/*****************************************************************************
*                                GLOBAL DATA                                 *
*****************************************************************************/

/* memory management for fixed-size blocks */
void *MPIR_shandles;		/* sbcnst MPIR_SHANDLES */
void *MPIR_rhandles;		/* sbcnst MPIR_RHANDLES */
void *MPIR_dtes;		/* sbcnst datatype elements */
void *MPIR_qels;		/* sbcnst queue elements */
void *MPIR_fdtels;		/* sbcnst flat datatype elements */

/* pre-assigned datatypes */
MPIR_DATATYPE *MPIR_int_dte;
MPIR_DATATYPE *MPIR_float_dte;
MPIR_DATATYPE *MPIR_double_dte;
MPIR_DATATYPE *MPIR_long_dte;
MPIR_DATATYPE *MPIR_short_dte;
MPIR_DATATYPE *MPIR_char_dte;
MPIR_DATATYPE *MPIR_byte_dte;

/* queues */
MPIR_QHDR posted_sends;
MPIR_QHDR posted_recvs;
MPIR_QHDR unexpected_recvs;

/* MPIR routines */
extern MPIR_GROUP *MPIR_CreateGroup();

#endif
